# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Parse inference model predictions to triton inference responses
per model type.
"""

from functools import wraps
from typing import Callable

import numpy as np

try:
  import triton_python_backend_utils as pb_utils
except ModuleNotFoundError:
  pass


def visual_detector(func: Callable):
  """
  Visual detector type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray):
    """
    Format predictions and return clarifai compatible output.
    """
    out_bboxes = []
    out_labels = []
    out_scores = []
    for item in input_data:
      preds = func(self, item)
      out_bboxes.append(preds.predicted_bboxes)
      out_labels.append(preds.predicted_labels)
      out_scores.append(preds.predicted_scores)

    if len(out_bboxes) < 1 or len(out_labels) < 1:
      out_tensor_bboxes = pb_utils.Tensor("predicted_bboxes", np.zeros((0, 4), dtype=np.float32))
      out_tensor_labels = pb_utils.Tensor("predicted_labels", np.zeros((0, 1), dtype=np.int32))
      out_tensor_scores = pb_utils.Tensor("predicted_scores", np.zeros((0, 1), dtype=np.float32))
    else:
      out_tensor_bboxes = pb_utils.Tensor("predicted_bboxes",
                                          np.asarray(out_bboxes, dtype=np.float32))
      out_tensor_labels = pb_utils.Tensor("predicted_labels",
                                          np.asarray(out_labels, dtype=np.int32))
      out_tensor_scores = pb_utils.Tensor("predicted_scores",
                                          np.asarray(out_scores, dtype=np.float32))

    inference_response = pb_utils.InferenceResponse(
        output_tensors=[out_tensor_bboxes, out_tensor_labels, out_tensor_scores])

    return inference_response

  return parse_predictions


def visual_classifier(func: Callable):
  """
  Visual classifier type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray):
    """
    Format predictions and return clarifai compatible output.
    """
    out_scores = []
    for item in input_data:
      preds = func(self, item)
      out_scores.append(preds.predicted_scores)

    out_tensor_scores = pb_utils.Tensor("softmax_predictions",
                                        np.asarray(out_scores, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor_scores])

    return inference_response

  return parse_predictions


def text_classifier(func: Callable):
  """
  Text classifier type output parser.
  """

  @wraps(func)
  def parse_predictions(self, input_data: np.ndarray):
    """
    Format predictions and return clarifai compatible output.
    """
    out_scores = []
    input_data = [in_elem[0].decode() for in_elem in input_data]
    for item in input_data:
      preds = func(self, item)
      out_scores.append(preds.predicted_scores)

    out_tensor_scores = pb_utils.Tensor("softmax_predictions",
                                        np.asarray(out_scores, dtype=np.float32))
    inference_response = pb_utils.InferenceResponse(output_tensors=[out_tensor_scores])

    return inference_response

  return parse_predictions



import glob
import json
import math
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from typing import DefaultDict, Dict, List

import cv2
from tqdm import tqdm

from ..features import VisualDetectionFeatures


class xviewDetectionDataset:
  """xview Image Detection Dataset"""

  xview_concepts = [
      'Fixed-wing Aircraft', 'Small Aircraft', 'Cargo Plane', 'Helicopter', 'Passenger Vehicle',
      'Small Car', 'Bus', 'Pickup Truck', 'Utility Truck', 'Truck', 'Cargo Truck', 'Truck w-Box',
      'Truck Tractor', 'Trailer', 'Truck w-Flatbed', 'Truck w-Liquid', 'Crane Truck',
      'Railway Vehicle', 'Passenger Car', 'Cargo Car', 'Flat Car', 'Tank car', 'Locomotive',
      'Maritime Vessel', 'Motorboat', 'Sailboat', 'Tugboat', 'Barge', 'Fishing Vessel', 'Ferry',
      'Yacht', 'Container Ship', 'Oil Tanker', 'Engineering Vehicle', 'Tower crane',
      'Container Crane', 'Reach Stacker', 'Straddle Carrier', 'Mobile Crane', 'Dump Truck',
      'Haul Truck', 'Scraper-Tractor', 'Front loader-Bulldozer', 'Excavator', 'Cement Mixer',
      'Ground Grader', 'Hut-Tent', 'Shed', 'Building', 'Aircraft Hangar', 'Damaged Building',
      'Facility', 'Construction Site', 'Vehicle Lot', 'Helipad', 'Storage Tank',
      'Shipping container lot', 'Shipping Container', 'Pylon', 'Tower'
  ]

  def __init__(self, split: str = "train") -> None:
    """
        Initialize and Compress xview dataset.
        Args:
        split: "train"
        """

    self.data_dir = os.path.join(os.curdir, "data")
    self.img_dir = os.path.join(self.data_dir, "train_images")
    self.img_comp_dir = os.path.join(self.data_dir, "train_images_comp")
    self.label_file = os.path.join(self.data_dir, "xview_train.geojson")
    self.max_annotations_per_img = 700  # Max annotations per image(due to max grpc 10KB limit)

    self.preprocess()
    self.all_data = self.xview_data_parser()

  def compress_tiff(self, img_path: str) -> None:
    """Compress tiff image"""
    img_comp_path = os.path.join(self.img_comp_dir, os.path.basename(img_path))
    img_arr = cv2.imread(img_path)
    cv2.imwrite(
        img_comp_path, img_arr, params=(cv2.IMWRITE_TIFF_COMPRESSION, 8))  # 8: Adobe Deflate

  def preprocess(self):
    """
        Compress the tiff images to comply with clarifai grpc image encoding limit(<20MB)
        Uses ADOBE_DEFLATE compression algorithm
        """
    all_img_ids = glob.glob(os.path.join(self.img_dir, "*.tif"))

    if not os.path.exists(self.img_comp_dir):
      os.mkdir(self.img_comp_dir)

    num_workers = cpu_count()
    futures = []

    with ThreadPoolExecutor(max_workers=num_workers) as executor:
      with tqdm(total=len(all_img_ids), desc="compressing ...") as progress:

        for img_path in all_img_ids:
          future = executor.submit(self.compress_tiff, img_path)
          future.add_done_callback(lambda _: progress.update())
          futures.append(future)

        results = []
        for future in futures:
          result = future.result()
          results.append(result)

  def xview_classes2indices(self, classes: List) -> List:
    """ remap xview classes 11-94 to 0-59 """
    indices = [
        -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 1, 2, -1, 3, -1, 4, 5, 6, 7, 8, -1, 9, 10,
        11, 12, 13, 14, 15, -1, -1, 16, 17, 18, 19, 20, 21, 22, -1, 23, 24, 25, -1, 26, 27, -1, 28,
        -1, 29, 30, 31, 32, 33, 34, 35, 36, 37, -1, 38, 39, 40, 41, 42, 43, 44, 45, -1, -1, -1, -1,
        46, 47, 48, 49, -1, 50, 51, -1, 52, -1, -1, -1, 53, 54, -1, 55, -1, -1, 56, -1, 57, -1, 58,
        59
    ]
    return [indices[int(c)] for c in classes]

  def xview_indices2concepts(self, indices: List) -> List:
    """ remap classes to concept names """
    return [self.xview_concepts[i] for i in indices]

  def xview_data_parser(self) -> DefaultDict[str, Dict[List, List]]:
    """
        Parse geojson data into nested dict of imageid w.r.t bounding boxes, concepts
        """
    all_data = defaultdict(lambda: dict(bboxes=[], concepts=[]))

    with open(self.label_file) as f:
      geojson_data = json.loads(f.read())

    for feature in tqdm(
        geojson_data['features'], total=len(geojson_data['features']),
        desc="Parsing geojson data"):
      image_id = feature['properties']['image_id'].split(".")[0]
      xview_classid = feature['properties']['type_id']
      bbox = list(map(int, feature['properties']['bounds_imcoords'].split(",")))
      concept = self.xview_indices2concepts(self.xview_classes2indices([xview_classid]))
      geo_loc = feature['geometry']['coordinates']  # Long,Lat

      all_data[image_id]['bboxes'].append(bbox)
      all_data[image_id]['concepts'].append(concept[0])

    return all_data

  def dataloader(self):
    """
        Transform xview object detection data into clarifai proto compatible
        format for upload.
        Returns:
            VisualDetectionFeatures type generator.
        """
    img_ids = list(self.all_data.keys())

    #get annotations for each image id
    for _id in img_ids:
      image_path = os.path.join(self.img_comp_dir, _id + ".tif")
      if not os.path.exists(image_path):
        print(f"Invalid Image path... {image_path}")
        continue

      image_size_MB = os.path.getsize(image_path) / 1048576  #Skip compressed image > 20MB
      if math.ceil(image_size_MB) > 20:
        continue

      image_height, image_width = cv2.imread(image_path).shape[:2]
      annots = []
      class_names = []
      for bbox, concept in zip(self.all_data[_id]['bboxes'][:self.max_annotations_per_img],
                               self.all_data[_id]['concepts'][:self.max_annotations_per_img]):
        x_min = max(min(bbox[0] / image_width, 1.0), 0.0)  #left_col
        y_min = max(min(bbox[1] / image_height, 1.0), 0.0)  #top_row
        x_max = max(min(bbox[2] / image_width, 1.0), 0.0)  #right_col
        y_max = max(min(bbox[3] / image_height, 1.0), 0.0)  #bottom_row
        if (x_min >= x_max) or (y_min >= y_max):
          continue
        annots.append([round(x, 3) for x in [x_min, y_min, x_max, y_max]])
        class_names.append(concept)

      assert len(class_names) == len(annots), f"Num classes must match num bbox annotations\
            for a single image. Found {len(class_names)} classes and {len(annots)} bboxes."

      yield VisualDetectionFeatures(image_path, class_names, annots, id=_id)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton Model Config classes."""

from dataclasses import dataclass, field
from typing import List


@dataclass
class ClarifaiFieldsMap:
  """
  Triton Model Config base.
  Params:
  -------
  model_type

  Returns:
  --------
  ClarifaiFieldsMap
  """
  model_type: str
  input_fields_map: List = field(default_factory=list)
  output_fields_map: List = field(default_factory=list)

  def __post_init__(self):
    """
    Set mapping of clarifai in/output vs triton in/output
    """
    if self.model_type == "visual-detector":
      self.input_fields_map = {"image": "image"}
      self.output_fields_map = {
          "regions[...].region_info.bounding_box": "predicted_bboxes",
          "regions[...].data.concepts[...].id": "predicted_labels",
          "regions[...].data.concepts[...].value": "predicted_scores"
      }
    elif self.model_type == "visual-classifier":
      self.input_fields_map = {"image": "image"}
      self.output_fields_map = {"concepts": "softmax_predictions"}
    elif self.model_type == "text-classifier":
      self.input_fields_map = {"text": "text"}
      self.output_fields_map = {"concepts": "softmax_predictions"}



from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client import V2Stub


def concepts_generator(stub: V2Stub, user_id: str, app_id: str, page_size: int = 64):
  """
    Lists all the concept in the given userAppID user_id, app_id app.

    Args:
      stub: client stub.
      user_id: the user to list from.
      app_id: the app in the user_id account to list from.
      page_size: the pagination size to use while iterating.

    Returns:
      concepts: a list of Concept protos for all the concepts in the app.
    """
  userDataObject = resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)

  concept_success_status = {status_code_pb2.SUCCESS}

  page = 1
  while True:
    response = stub.ListConcepts(
        service_pb2.ListConceptsRequest(user_app_id=userDataObject, page=page, per_page=page_size),
    )

    if response.status.code not in concept_success_status:
      raise Exception("ListConcepts failed with response %r" % response)
    if len(response.concepts) == 0:
      break
    else:
      for item in response.concepts:
        yield item
    page += 1



from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.client import V2Stub


def installed_module_versions_generator(stub: V2Stub,
                                        user_id: str,
                                        app_id: str,
                                        page_size: int = 64):
  """
  Lists all the installed module versions in the given userAppID user_id, app_id app.

  Args:
    stub: client stub.
    user_id: the user to list from.
    app_id: the app in the user_id account to list from.
    page_size: the pagination size to use while iterating.

  Returns:
    imvs: a list of InstalledModuleVersion protos for all the installed modules versions in the app.
  """
  userDataObject = resources_pb2.UserAppIDSet(user_id=user_id, app_id=app_id)

  imv_success_status = {status_code_pb2.SUCCESS}

  page = 1
  while True:
    response = stub.ListInstalledModuleVersions(
        service_pb2.ListInstalledModuleVersionsRequest(
            user_app_id=userDataObject, page=page, per_page=page_size),)

    if response.status.code not in imv_success_status:
      raise Exception("ListInstalledModuleVersions failed with response %r" % response)
    for item in response.installed_module_versions:
      yield item
    page += 1
    # if we don't get a full page back we know we're done.
    if len(response.installed_module_versions) < page_size:
      break



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton model repository generation commandline interface."""

import argparse

from ..constants import MODEL_TYPES
from ..model_config.triton_config import TritonModelConfig
from ..pb_model_repository import TritonModelRepository


def model_upload_init():
  """
  Clarifai triton model upload commandline tool.
  """
  parser = argparse.ArgumentParser(description=__doc__)
  # TritonModelConfig args
  parser.add_argument("--model_name", type=str, required=True, help="Inference Model Name")
  parser.add_argument(
      "--model_version",
      type=str,
      default="1",
      required=False,
      help="Triton inference model version name. 1 stands for version 1. \
    Leave as default value (Recommended).")
  parser.add_argument(
      "--model_type",
      type=str,
      choices=MODEL_TYPES,
      required=True,
      help=f"Clarifai supported model types.\n Model-types-map: {MODEL_TYPES}",
  )
  parser.add_argument(
      "--repo_dir",
      type=str,
      default=".",  #curdir
      required=True,
      help="Directory to create triton repository.")

  args = parser.parse_args()

  model_config = TritonModelConfig(
      model_name=args.model_name, model_version="1", model_type=args.model_type)

  triton_repo = TritonModelRepository(model_config)
  triton_repo.build_repository(args.repo_dir)


if __name__ == "__main__":
  model_upload_init()



# This file contains boilerplate code to allow users write their model
# inference code that will then interact with the Triton Inference Server
# Python backend to serve end user requests.
# The module name, module path, class name & get_predictions() method names MUST be maintained as is
# but other methods may be added within the class as deemed fit provided
# they are invoked within the main get_predictions() inference method
# if they play a role in any step of model inference
"""User model inference script."""

import os
from pathlib import Path
from typing import Callable

import torch
from scipy.special import softmax
from transformers import ViTFeatureExtractor, ViTForImageClassification

from clarifai.models.model_serving.models.model_types import visual_classifier
from clarifai.models.model_serving.models.output import ClassifierOutput


class InferenceModel:
  """User model inference class."""

  def __init__(self) -> None:
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    self.base_path: Path = os.path.dirname(__file__)
    self.huggingface_model_path: Path = os.path.join(self.base_path, "vit-age-classifier")
    self.transforms = ViTFeatureExtractor.from_pretrained(self.huggingface_model_path)
    self.model: Callable = ViTForImageClassification.from_pretrained(self.huggingface_model_path)
    self.device = "cuda:0" if torch.cuda.is_available() else "cpu"

  @visual_classifier
  def get_predictions(self, input_data) -> ClassifierOutput:
    """
    Main model inference method.

    Args:
    -----
      input_data: A single input data item to predict on.
        Input data can be an image or text, etc depending on the model type.

    Returns:
    --------
      One of the clarifai.models.model_serving.models.output types. Refer to the README/docs
    """
    # Transform image and pass it to the model
    inputs = self.transforms(input_data, return_tensors='pt')
    output = self.model(**inputs)
    pred_scores = softmax(
        output[0][0].detach().numpy())  # alt: softmax(output.logits[0].detach().numpy())

    return ClassifierOutput(predicted_scores=pred_scores)



import csv
import os
import sys
from itertools import chain

import pandas as pd
from sklearn.model_selection import train_test_split


class DataPreprocessor:

  def __init__(self, filename, multi_val, separator):
    self.filename = filename
    self.multi_val = multi_val
    self.separator = separator
    self.df = pd.read_csv(filename)

  def process_data(self):
    text_col = self.get_column("Enter the number of the column that contains the text: ")
    label_col = self.get_column(
        "Enter the number of the column that contains the labels: ", exclude=text_col)

    self.split_values(label_col)
    self.df[label_col] = self.df[label_col].apply(lambda x: [x] if isinstance(x, str) else x)

    # Use chain.from_iterable to expand the multi-values if applicable
    unique_labels = list(set(chain.from_iterable(self.df[label_col].values)))

    print(
        "\nThe following unqiue labels have been found in the '{}' column and will be used in the dataset:".
        format(label_col))
    for i, label in enumerate(unique_labels, start=1):
      print('{}) {}'.format(i, label))

    self.convert_for_classification(text_col, label_col, unique_labels)

  def get_column(self, prompt, exclude=None):
    available_columns = self.df.columns.drop(exclude) if exclude else self.df.columns
    if len(available_columns) == 1:
      print(f'\nThe column named \'{available_columns[0]}\' will be used as the labels column.')
      return available_columns[0]
    else:
      for i, col in enumerate(available_columns):
        print(f'{i+1}) {col}')
      col_index = int(input(prompt)) - 1
      return available_columns[col_index]

  def split_values(self, label_col):
    if self.multi_val.lower() == 'y':
      self.df[label_col] = self.df[label_col].apply(
          lambda x: str(x).split(self.separator) if not isinstance(x, float) else x)

  def convert_for_classification(self, text_col, label_col, unique_labels):
    # Binary classification
    if len(unique_labels) == 2:
      print("Converting the CSV to be used with binary classification")
      self.df['input.data.text.raw'] = self.df[text_col]
      self.df['input.data.concepts[0].id'] = label_col
      self.df['input.data.concepts[0].value'] = self.df[label_col].apply(
          lambda x: 1 if unique_labels[0] in x else 0)
      self.df = self.df[[
          'input.data.text.raw', 'input.data.concepts[0].id', 'input.data.concepts[0].value'
      ]]

    # Multi-class classification
    else:
      print("Converting the CSV to be used with multi-class classification")
      self.df['input.data.text.raw'] = self.df[text_col].apply(
          lambda x: x[0] if isinstance(x, list) else x)
      for i in range(len(unique_labels)):
        self.df[f'input.data.concepts[{i}].id'] = self.df[label_col].apply(
            lambda x: unique_labels[i] if unique_labels[i] in x else '')
        self.df[f'input.data.concepts[{i}].value'] = self.df[label_col].apply(
            lambda x: 1 if unique_labels[i] in x else '')

      self.df = self.df[['input.data.text.raw'] +
                        [f'input.data.concepts[{i}].id' for i in range(len(unique_labels))] +
                        [f'input.data.concepts[{i}].value' for i in range(len(unique_labels))]]

      # Reorder the columns
      cols = self.df.columns.tolist()
      new_cols = cols[:1]  # The first column 'input.data.text.raw'
      pairs = [[cols[i], cols[i + len(unique_labels)]] for i in range(1, len(unique_labels) + 1)]
      for pair in pairs:
        new_cols.extend(pair)
      self.df = self.df[new_cols]

      # Remove special characters from column names
      self.df.columns = self.df.columns.str.replace("^[\[]|[\]]$", "", regex=True)


class DatasetSplitter:

  def __init__(self, df, split_dataset, shuffle_dataset, seed=555):
    self.df = df
    self.split_dataset = split_dataset
    self.shuffle_dataset = shuffle_dataset
    self.seed = seed if seed != '' else 555

  def split_and_save(self, filename_base):
    if self.split_dataset.lower() == 'y':
      split_type = self.get_split_type()

      if split_type == 1:
        train_pct = self.get_percentage(
            'What percentage of the dataset should be used for training? Enter a number between 1 and 99: ',
            99)
        test_pct = 100 - train_pct
        print(f'Data will be split {train_pct}% train, {test_pct}% test')  # Added print statement
      elif split_type == 2:
        train_pct = self.get_percentage(
            'What percentage of the dataset should be used for training? Enter a number between 1 and 98: ',
            98)
        max_val_pct = 99 - train_pct  # Max percentage for validation is now reduced by 1
        val_pct = self.get_percentage(
            f'What percentage of the dataset should be used for validation? Enter a number between 1 and {max_val_pct}: ',
            max_val_pct)
        test_pct = 100 - train_pct - val_pct
        print(f'Data will be split {train_pct}% train, {val_pct}% validation, {test_pct}% test'
             )  # Added print statement

      train_df, test_df = train_test_split(
          self.df,
          test_size=test_pct / 100,
          random_state=self.seed,
          shuffle=self.shuffle_dataset.lower() == 'y')
      train_df.to_csv(filename_base + '-train.csv', index=False, quoting=csv.QUOTE_MINIMAL)
      test_df.to_csv(filename_base + '-test.csv', index=False, quoting=csv.QUOTE_MINIMAL)

      if split_type == 2:
        train_df, val_df = train_test_split(
            train_df,
            test_size=val_pct / 100,
            random_state=self.seed,
            shuffle=self.shuffle_dataset.lower() == 'y')
        val_df.to_csv(filename_base + '-validation.csv', index=False, quoting=csv.QUOTE_MINIMAL)
    else:
      self.df.to_csv(filename_base + '.csv', index=False, quoting=csv.QUOTE_MINIMAL)

  def get_split_type(self):
    split_type = int(
        input(
            'How would you like to split the dataset?\n1) Train and test datasets\n2) Train, validate, and test datasets\n'
        ))
    while split_type not in [1, 2]:
      split_type = int(
          input(
              'Invalid option. Enter 1 for "Train and test" or 2 for "Train, validate, and test": '
          ))
    return split_type

  def get_percentage(self, prompt, max_pct):
    pct = int(input(prompt))
    while not 1 <= pct <= max_pct:
      pct = int(input(f'Invalid input. Please enter a number between 1 and {max_pct}: '))
    return pct


def main():
  filename = sys.argv[1]
  multi_val = input('Do any columns have multiple values? (y/[n]) ')
  separator = input('Enter the separator: ') if multi_val.lower() == 'y' else None

  preprocessor = DataPreprocessor(filename, multi_val, separator)
  preprocessor.process_data()

  split_dataset = input('Would you like to split this dataset? (y/[n]) ')
  shuffle_dataset = 'n'
  seed = '555'

  if split_dataset.lower() == 'y':
    shuffle_dataset = input('Would you like to shuffle the dataset before splitting? (y/[n]) ')
    if shuffle_dataset.lower() == 'y':
      seed = input('Enter a seed integer or hit enter to use the default [555]: ')

  splitter = DatasetSplitter(preprocessor.df, split_dataset, shuffle_dataset, seed)
  splitter.split_and_save(os.path.splitext(filename)[0] + '-clarifai')
  print("Done!")


if __name__ == "__main__":
  main()



#! COCO 2017 detection dataset

import os
import zipfile
from glob import glob

import cv2
import requests
from pycocotools.coco import COCO
from tqdm import tqdm

from ..features import VisualDetectionFeatures


class COCODetectionDataset:
  """COCO 2017 Image Detection Dataset."""

  def __init__(self, split: str = "train"):
    """
    Initialize coco dataset.
    Args:
      filenames: the coco zip filenames: Dict[str, str] to be downloaded if download=True,
      data_dir: the local coco dataset directory.
      split: "train" or "val"
    """
    self.filenames = {
        "train": "train2017.zip",
        "val": "val2017.zip",
        "annotations": "annotations_trainval2017.zip"
    }
    self.split = split
    self.url = "http://images.cocodataset.org/zips/"  # coco base image-zip url
    self.data_dir = os.path.join(os.curdir, "data")  # data storage directory
    self.extracted_coco_dirs = {"train": None, "val": None, "annotations": None}

  def coco_download(self, save_dir):
    """Download coco dataset."""
    if not os.path.exists(save_dir):
      os.mkdir(save_dir)

    #check if train*, val* and annotation* dirs exist
    #so that the coco2017 data isn't downloaded
    for key, filename in self.filenames.items():
      existing_files = glob(f"{save_dir}/{key}*")
      if existing_files:
        print(f"{key} dataset already downloded and extracted")
        continue

      print("-" * 80)
      print(f"Downloading {filename}")
      print("-" * 80)

      if "annotations" in filename:
        self.url = "http://images.cocodataset.org/annotations/"

      response = requests.get(self.url + filename, stream=True)
      response.raise_for_status()
      with open(os.path.join(save_dir, filename), "wb") as _file:
        for chunk in tqdm(response.iter_content(chunk_size=5124000)):
          if chunk:
            _file.write(chunk)
      print("Coco data download complete...")

      #extract files
      zf = zipfile.ZipFile(os.path.join(save_dir, filename))
      print(f" Extracting {filename} file")
      zf.extractall(path=save_dir)
      # Delete coco zip
      print(f" Deleting {filename}")
      os.remove(path=os.path.join(save_dir, filename))

  def dataloader(self):
    """
    Transform coco object detection data into clarifai proto compatible
    format for upload.
    Returns:
      VisualDetectionFeatures type generator.
    """
    if isinstance(self.filenames, dict) and len(self.filenames) == 3:
      self.coco_download(self.data_dir)
      self.extracted_coco_dirs["train"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "train" in i][0]
      self.extracted_coco_dirs["val"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "val" in i][0]

      self.extracted_coco_dirs["annotations"] = [os.path.join(self.data_dir, i) \
      for i in os.listdir(self.data_dir) if "annotations" in i][0]
    else:
      raise Exception(f"`filenames` must be a dict of atleast 2 coco zip file names; \
      train, val and annotations. Found {len(self.filenames)} items instead.")

    annot_file = glob(self.extracted_coco_dirs["annotations"] + "/" +\
     f"instances_{self.split}*")[0]
    coco = COCO(annot_file)
    categories = coco.loadCats(coco.getCatIds())
    cat_id_map = {category["id"]: category["name"] for category in categories}
    cat_img_ids = {}
    for cat_id in list(cat_id_map.keys()):
      cat_img_ids[cat_id] = coco.getImgIds(catIds=[cat_id])

    img_ids = []
    for i in list(cat_img_ids.values()):
      img_ids.extend(i)

    #get annotations for each image id
    for _id in set(img_ids):
      annots = []  # bboxes
      class_names = []
      labels = [i for i in list(filter(lambda x: _id in cat_img_ids[x], cat_img_ids))]
      image_path = glob(self.extracted_coco_dirs[self.split]+"/"+\
      f"{str(_id).zfill(12)}*")[0]

      image_height, image_width = cv2.imread(image_path).shape[:2]
      for cat_id in labels:
        annot_ids = coco.getAnnIds(imgIds=_id, catIds=[cat_id])
        if len(annot_ids) > 0:
          img_annotations = coco.loadAnns(annot_ids)
          for ann in img_annotations:
            class_names.append(cat_id_map[cat_id])
            x_min = ann['bbox'][0] / image_width  #left_col
            y_min = ann['bbox'][1] / image_height  #top_row
            x_max = (ann['bbox'][0] + ann['bbox'][2]) / image_width  #right_col
            y_max = (ann['bbox'][1] + ann['bbox'][3]) / image_height  #bottom_row
            annots.append([x_min, y_min, x_max, y_max])
        else:  # if no annotations for given image_id-cat_id pair
          continue
      assert len(class_names) == len(annots), f"Num classes must match num bbox annotations\
      for a single image. Found {len(class_names)} classes and {len(annots)} bboxes."

      yield VisualDetectionFeatures(image_path, class_names, annots, id=_id)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Parse & Serialize TritonModelConfig objects into proto format.
"""

import os
from pathlib import Path
from typing import Type

from google.protobuf.text_format import MessageToString
from tritonclient.grpc import model_config_pb2

from .triton_config import TritonModelConfig


class Serializer:
  """
  Serialize TritonModelConfig type object.
  """

  def __init__(self, model_config: Type[TritonModelConfig]) -> None:
    self.model_config = model_config  #python dataclass config
    self.config_proto = model_config_pb2.ModelConfig()  #holds parsed python config

    self._set_all_fields()

  def _set_input(self) -> None:
    """
    Parse InputConfig object to proto.
    """
    input_config = self.config_proto.input.add()
    for in_field in self.model_config.input:
      for key, value in in_field.__dict__.items():
        try:
          setattr(input_config, key, value)
        except AttributeError:
          field = getattr(input_config, key)
          if isinstance(value, list):
            field.extend(value)
          else:
            field.extend([value])
    return

  def _set_output(self) -> None:
    """
    Parse OutputConfig object to proto.
    """
    # loop over output dataclass list
    for out_field in self.model_config.output:
      output_config = self.config_proto.output.add()
      for key, value in out_field.__dict__.items():
        try:
          setattr(output_config, key, value)
        except AttributeError:  #Proto Repeated Field assignment not allowed
          field = getattr(output_config, key)
          if isinstance(value, list):
            field.extend(value)
          else:
            field.extend([value])
    return

  def _set_instance_group(self) -> None:
    """
    Parse triton model instance group settings to proto.
    """
    instance = self.config_proto.instance_group.add()
    for field_name, value in self.model_config.instance_group.__dict__.items():
      try:
        setattr(instance, field_name, value)
      except AttributeError:
        continue
    return

  def _set_batch_info(self) -> model_config_pb2.ModelDynamicBatching:
    """
    Parse triton model dynamic batching settings to proto.
    """
    dbatch_msg = model_config_pb2.ModelDynamicBatching()
    for key, value in self.model_config.dynamic_batching.__dict__.items():
      try:
        setattr(dbatch_msg, key, value)
      except AttributeError:  #Proto Repeated Field assignment not allowed
        field = getattr(dbatch_msg, key)
        if isinstance(value, list):
          field.extend(value)
        else:
          field.extend([value])

    return dbatch_msg

  def _set_all_fields(self) -> None:
    """
    Set all config fields.
    """
    self.config_proto.name = self.model_config.model_name
    self.config_proto.backend = self.model_config.backend
    self.config_proto.max_batch_size = self.model_config.max_batch_size
    self._set_input()
    self._set_output()
    self._set_instance_group()
    dynamic_batch_msg = self._set_batch_info()
    self.config_proto.dynamic_batching.CopyFrom(dynamic_batch_msg)

  @property
  def get_config(self) -> model_config_pb2.ModelConfig:
    """
    Return model config proto.
    """
    return self.config_proto

  def to_file(self, save_dir: Path) -> None:
    """
    Serialize all triton config parameters and save output
    to file.
    Args:
    -----
    save_dir: Directory where to save resultant config.pbtxt file.
          Defaults to the current working dir.
    """
    msg_string = MessageToString(self.config_proto)

    with open(os.path.join(save_dir, "config.pbtxt"), "w") as cfile:
      cfile.write(msg_string)



from collections import defaultdict
from typing import Iterator, List, Tuple

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct


class ClarifaiDataset:
  """
  Clarifai datasets base class.
  """

  def __init__(self, datagen_object: Iterator, dataset_id: str, split: str) -> None:
    self.datagen_object = datagen_object
    self.dataset_id = dataset_id
    self.split = split
    self.input_ids = []
    self._all_input_protos = {}
    self._all_annotation_protos = defaultdict(list)

  def __len__(self) -> int:
    """
    Get size of all input protos
    """
    return len(self._all_input_protos)

  def _to_list(self, input_protos: Iterator) -> List:
    """
    Parse protos iterator to list.
    """
    return list(input_protos)

  def create_input_protos(self, image_path: str, label: str, input_id: str, dataset_id: str,
                          metadata: Struct) -> resources_pb2.Input:
    """
    Create input protos for each image, label input pair.
    Args:
    	`image_path`: path to image.
    	`label`: image label
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`: input metadata
    Returns:
    	An input proto representing a single row input
    """
    raise NotImplementedError()

  def _extract_protos(self) -> None:
    """
    Create input image protos for each data generator item.
    """
    raise NotImplementedError()

  def get_protos(self, input_ids: List[str]
                ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
    """
    Get input and annotation protos based on input_ids.
    Args:
      `input_ids`: List of input IDs to retrieve the protos for.
    Returns:
      Input and Annotation proto iterators for the specified input IDs.
    """
    input_protos = [self._all_input_protos.get(input_id) for input_id in input_ids]
    annotation_protos = []
    if len(self._all_annotation_protos) > 0:
      annotation_protos = [self._annotation_protos.get(input_id) for input_id in input_ids]
      annotation_protos = [
          ann_proto for ann_protos in annotation_protos for ann_proto in ann_protos
      ]

    return input_protos, annotation_protos


class Chunker:
  """
  Split an input sequence into small chunks.
  """

  def __init__(self, seq: List, size: int) -> None:
    self.seq = seq
    self.size = size

  def chunk(self) -> List[List]:
    """
    Chunk input sequence.
    """
    return [self.seq[pos:pos + self.size] for pos in range(0, len(self.seq), self.size)]



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Triton python backend inference model controller.
"""

import inspect
import os
from pathlib import Path
from typing import Callable, Type

from .model_config.serializer import Serializer
from .model_config.triton_config import TritonModelConfig
from .models import inference, pb_model


class TritonModelRepository:
  """
  Triton Python BE Model Repository Generator.
  """

  def __init__(self, model_config: Type[TritonModelConfig]):
    self.model_config = model_config
    self.config_proto = Serializer(model_config)

  def _module_to_file(self, module_name: Callable, filename: str, destination_dir: str) -> None:
    """
    Write Python Module to file.

    Args:
    -----
    module_name: Python module name to write to file
    filename: Name of the file to write to destination_dir
    destination_dir: Directory to save the generated triton model file.

    Returns:
    --------
    None
    """
    module_path: Path = os.path.join(destination_dir, filename)
    source_code = inspect.getsource(module_name)
    with open(module_path, "w") as pb_model:
      pb_model.write(source_code)

  def build_repository(self, repository_dir: Path = os.curdir):
    """
    Generate Triton Model Repository.

    Args:
    -----
    repository_dir: Directory to create triton model repository

    Returns:
    --------
    None
    """
    model_repository = self.model_config.model_name
    model_version = self.model_config.model_version
    repository_path = os.path.join(repository_dir, model_repository)
    model_version_path = os.path.join(repository_path, model_version)

    if not os.path.isdir(repository_path):
      os.mkdir(repository_path)
      self.config_proto.to_file(repository_path)
      for out_field in self.model_config.output:
        #predicted int labels must have corresponding names in file
        if hasattr(out_field, "label_filename"):
          with open(os.path.join(repository_path, "labels.txt"), "w"):
            pass
        else:
          continue
      # gen requirements & conda yaml
      with open(os.path.join(repository_path, "requirements.txt"), "w") as f:
        f.write("clarifai>9.5.3\ntritonclient[all]")  # for model upload utils
      with open(os.path.join(repository_path, "triton_conda.yaml"), "w") as conda_env:
        conda_env.write("name: triton_conda-cp3.8-torch1.13.1-19f97078")

    if not os.path.isdir(model_version_path):
      os.mkdir(model_version_path)
    if not os.path.exists(os.path.join(model_version_path, "__init__.py")):
      with open(os.path.join(model_version_path, "__init__.py"), "w"):
        pass
    # generate model.py & inference.py modules
    self._module_to_file(pb_model, filename="model.py", destination_dir=model_version_path)
    self._module_to_file(inference, filename="inference.py", destination_dir=model_version_path)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Output Predictions format for different model types.
"""

from dataclasses import dataclass
import numpy as np


@dataclass
class VisualDetectorOutput:
  predicted_bboxes: np.ndarray
  predicted_labels: np.ndarray
  predicted_scores: np.ndarray

  def __post_init__(self):
    """
    Validate input upon initialization.
    """
    assert self.predicted_bboxes.ndim == self.predicted_labels.ndim == \
      self.predicted_scores.ndim==2, f"All predictions must be 2-dimensional, \
        Got bbox-dims: {self.predicted_bboxes.ndim}, label-dims: {self.predicted_labels.ndim}, \
          scores-dims: {self.predicted_scores.ndim} instead."
    assert self.predicted_bboxes.shape[0] == self.predicted_labels.shape[0] == \
      self.predicted_scores.shape[0], f"The Number of predicted bounding boxes, \
        predicted labels and predicted scores MUST match. Got {len(self.predicted_bboxes)}, \
          {self.predicted_labels.shape[0]}, {self.predicted_scores.shape[0]} instead."

    if len(self.predicted_labels) > 0:
      assert self.predicted_bboxes.shape[1] == 4, f"Box coordinates must have a length of 4."
      assert np.all(np.logical_and(0 <= self.predicted_bboxes, self.predicted_bboxes <= 1)), \
       "Bounding box coordinates must be between 0 and 1"


@dataclass
class ClassifierOutput:
  """
  Takes model softmax predictions
  """
  predicted_scores: np.ndarray

  # the index of each predicted score as returned by the model must correspond
  # to the predicted label index in the labels.txt file

  def __post_init__(self):
    """
    Validate input upon initialization.
    """
    assert self.predicted_scores.ndim == 1, \
      f"All predictions must be 1-dimensional, Got scores-dims: {self.predicted_scores.ndim} instead."



#! Pascal VOC Object Detection dataset.

import os

try:
  import xml.etree.cElementTree as ET
except ImportError:
  import xml.etree.ElementTree as ET

from clarifai.data_upload.datasets.features import VisualDetectionFeatures


class VOCDetectionDataset:
  """Pascal VOC Image Detection Dataset.
  url: http://host.robots.ox.ac.uk/pascal/VOC/voc2012/index.html
  """

  voc_concepts = [
      'aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus', 'car', 'cat', 'chair', 'cow',
      'diningtable', 'dog', 'horse', 'motorbike', 'person', 'pottedplant', 'sheep', 'sofa',
      'train', 'tvmonitor'
  ]

  def __init__(self, split: str = "train") -> None:
    """
    Inititalize dataset params.
    Args:
      split: "train" or "test"
    """

    self.split = split
    self.image_dir = {"train": os.path.join(os.path.dirname(__file__), "images")}
    self.annotations_dir = {"train": os.path.join(os.path.dirname(__file__), "annotations")}

  def dataloader(self):
    """
    Transform Pascal VOC detection dataset into clarifai proto compatible
    format to uplaod
    Returns:
      VisualDetectionFeatures type generator.
    """

    all_imgs = os.listdir(self.image_dir[self.split])
    img_ids = [img_filename.split('.')[0] for img_filename in all_imgs]

    for _id in img_ids:
      image_path = os.path.join(self.image_dir[self.split], _id + ".jpg")
      annot_path = os.path.join(self.annotations_dir[self.split], _id + ".xml")

      root = ET.parse(annot_path).getroot()
      size = root.find('size')
      width = float(size.find('width').text)
      height = float(size.find('height').text)

      annots = []
      class_names = []
      for obj in root.iter('object'):
        concept = obj.find('name').text.strip().lower()
        if concept not in self.voc_concepts:
          continue
        xml_box = obj.find('bndbox')
        #Making bounding box to be 0-1
        x_min = max(min((float(xml_box.find('xmin').text) - 1) / width, 1.0), 0.0)  #left_col
        y_min = max(min((float(xml_box.find('ymin').text) - 1) / height, 1.0), 0.0)  #top_row
        x_max = max(min((float(xml_box.find('xmax').text) - 1) / width, 1.0), 0.0)  #right_col
        y_max = max(min((float(xml_box.find('ymax').text) - 1) / height, 1.0), 0.0)  #bottom_row

        if (x_min >= x_max) or (y_min >= y_max):
          continue
        annots.append([x_min, y_min, x_max, y_max])
        class_names.append(concept)

      assert len(class_names) == len(annots), f"Num classes must match num bbox annotations\
            for a single image. Found {len(class_names)} classes and {len(annots)} bboxes."

      yield VisualDetectionFeatures(image_path, class_names, annots, id=_id)



import os
from typing import Iterator, List, Union

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct
from tqdm import tqdm

from .base import ClarifaiDataset


class VisualClassificationDataset(ClarifaiDataset):

  def __init__(self, datagen_object: Iterator, dataset_id: str, split: str) -> None:
    super().__init__(datagen_object, dataset_id, split)
    self._extract_protos()

  def create_input_protos(self, image_path: str, labels: List[Union[str, int]], input_id: str,
                          dataset_id: str, geo_info: Union[List[float], None],
                          metadata: Struct) -> resources_pb2.Input:
    """
    Create input protos for each image, label input pair.
    Args:
      `image_path`: image path.
      `labels`: image label(s)
      `input_id: unique input id
      `dataset_id`: Clarifai dataset id
      `geo_info`: image longitude, latitude info
      `metadata`: image metadata
    Returns:
      An input proto representing a single row input
    """
    geo_pb = resources_pb2.Geo(geo_point=resources_pb2.GeoPoint(
        longitude=geo_info[0], latitude=geo_info[1])) if geo_info is not None else None

    input_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            image=resources_pb2.Image(base64=open(image_path, 'rb').read(),),
            geo=geo_pb,
            concepts=[
                resources_pb2.Concept(
                  id=f"id-{''.join(_label.split(' '))}", name=_label, value=1.)\
                for _label in labels
            ],
            metadata=metadata))

    return input_proto

  def _extract_protos(self) -> None:
    """
    Create input image protos for each data generator item.
    """
    for i, item in tqdm(enumerate(self.datagen_object), desc="Creating input protos..."):
      metadata = Struct()
      image_path = item.image_path
      label = item.label if isinstance(item.label, list) else [item.label]  # clarifai concept
      input_id = f"{self.dataset_id}-{self.split}-{i}" if item.id is None else f"{self.split}-{str(item.id)}"
      geo_info = item.geo_info
      metadata.update({"filename": os.path.basename(image_path), "split": self.split})

      self.input_ids.append(input_id)
      input_proto = self.create_input_protos(image_path, label, input_id, self.dataset_id,
                                             geo_info, metadata)
      self._all_input_protos[input_id] = input_proto


class VisualDetectionDataset(ClarifaiDataset):
  """
  Visual detection dataset proto class.
  """

  def __init__(self, datagen_object: Iterator, dataset_id: str, split: str) -> None:
    super().__init__(datagen_object, dataset_id, split)
    self._extract_protos()

  def create_input_protos(self, image_path: str, input_id: str, dataset_id: str,
                          geo_info: Union[List[float], None],
                          metadata: Struct) -> resources_pb2.Input:
    """
    Create input protos for each image, label input pair.
    Args:
      `image_path`: file path to image
      `input_id: unique input id
      `dataset_id`: Clarifai dataset id
      `geo_info`: image longitude, latitude info
      `metadata`: image metadata
    Returns:
      An input proto representing a single row input
    """
    geo_pb = resources_pb2.Geo(geo_point=resources_pb2.GeoPoint(
        longitude=geo_info[0], latitude=geo_info[1])) if geo_info is not None else None
    input_image_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            image=resources_pb2.Image(base64=open(image_path, 'rb').read(),),
            geo=geo_pb,
            metadata=metadata))

    return input_image_proto

  def create_annotation_proto(self, label: str, annotations: List, input_id: str,
                              dataset_id: str) -> resources_pb2.Annotation:
    """
    Create an input proto for each bounding box, label input pair.
    Args:
      `label`: annotation label
      `annotations`: a list of a single bbox's coordinates.
      `input_id: unique input id
      `dataset_id`: Clarifai dataset id
    Returns:
      An input proto representing a single image input
    """
    input_annot_proto = resources_pb2.Annotation(
        input_id=input_id,
        data=resources_pb2.Data(regions=[
            resources_pb2.Region(
                region_info=resources_pb2.RegionInfo(bounding_box=resources_pb2.BoundingBox(
                    # Annotations ordering: [xmin, ymin, xmax, ymax]
                    # top_row must be less than bottom row
                    # left_col must be less than right col
                    top_row=annotations[1],  #y_min
                    left_col=annotations[0],  #x_min
                    bottom_row=annotations[3],  #y_max
                    right_col=annotations[2]  #x_max
                )),
                data=resources_pb2.Data(concepts=[
                    resources_pb2.Concept(
                        id=f"id-{''.join(label.split(' '))}", name=label, value=1.)
                ]))
        ]))

    return input_annot_proto

  def _extract_protos(self) -> None:
    """
    Create input image protos for each data generator item.
    """
    for i, item in tqdm(enumerate(self.datagen_object), desc="Creating input protos..."):
      metadata = Struct()
      image = item.image_path
      labels = item.classes  # list:[l1,...,ln]
      bboxes = item.bboxes  # [[xmin,ymin,xmax,ymax],...,[xmin,ymin,xmax,ymax]]
      input_id = f"{self.dataset_id}-{self.split}-{i}" if item.id is None else f"{self.split}-{str(item.id)}"
      metadata.update({"filename": os.path.basename(image), "split": self.split})
      geo_info = item.geo_info

      self.input_ids.append(input_id)
      input_image_proto = self.create_input_protos(image, input_id, self.dataset_id, geo_info,
                                                   metadata)
      self._all_input_protos[input_id] = input_image_proto

      # iter over bboxes and classes
      # one id could have more than one bbox and label
      for i in range(len(bboxes)):
        input_annot_proto = self.create_annotation_proto(labels[i], bboxes[i], input_id,
                                                         self.dataset_id)
        self._all_annotation_protos[input_id].append(input_annot_proto)


class VisualSegmentationDataset(ClarifaiDataset):
  """
  Visual segmentation dataset proto class.
  """

  def __init__(self, datagen_object: Iterator, dataset_id: str, split: str) -> None:
    super().__init__(datagen_object, dataset_id, split)
    self._extract_protos()

  def create_input_protos(self, image_path: str, input_id: str, dataset_id: str,
                          geo_info: Union[List[float], None],
                          metadata: Struct) -> resources_pb2.Input:
    """
    Create input protos for each image, label input pair.
    Args:
      `image_path`: absolute image file path
      `input_id: unique input id
      `dataset_id`: Clarifai dataset id
      `geo_info`: image longitude, latitude info
      `metadata`: image metadata
    Returns:
      An input proto representing a single input item
    """
    geo_pb = resources_pb2.Geo(geo_point=resources_pb2.GeoPoint(
        longitude=geo_info[0], latitude=geo_info[1])) if geo_info is not None else None
    input_image_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            image=resources_pb2.Image(base64=open(image_path, 'rb').read(),),
            geo=geo_pb,
            metadata=metadata))

    return input_image_proto

  def create_mask_proto(self, label: str, polygons: List[List[float]], input_id: str,
                        dataset_id: str) -> resources_pb2.Annotation:
    """
    Create an input mask proto for an input polygon/mask and label.
    Args:
      `label`: image label
      `polygons`: Polygon x,y points iterable
      `input_id: unique input id
      `dataset_id`: Clarifai dataset id
    Returns:
      An input proto corresponding to a single image
    """
    input_mask_proto = resources_pb2.Annotation(
        input_id=input_id,
        data=resources_pb2.Data(regions=[
            resources_pb2.Region(
                region_info=resources_pb2.RegionInfo(polygon=resources_pb2.Polygon(
                    points=[
                        resources_pb2.Point(
                            row=_point[1],  # row is y point
                            col=_point[0],  # col is x point
                            visibility="VISIBLE") for _point in polygons
                    ])),
                data=resources_pb2.Data(concepts=[
                    resources_pb2.Concept(
                        id=f"id-{''.join(label.split(' '))}", name=label, value=1.)
                ]))
        ]))

    return input_mask_proto

  def _extract_protos(self) -> None:
    """
    Create input image and annotation protos for each data generator item.
    """
    for i, item in tqdm(enumerate(self.datagen_object), desc="Creating input protos..."):
      metadata = Struct()
      image = item.image_path  # image path
      labels = item.classes  # list of class labels
      _polygons = item.polygons  # list of polygons: [[[x,y],...,[x,y]],...]
      input_id = f"{self.dataset_id}-{self.split}-{i}" if item.id is None else f"{self.split}-{str(item.id)}"
      metadata.update({"filename": os.path.basename(image), "split": self.split})
      geo_info = item.geo_info

      self.input_ids.append(input_id)
      input_image_proto = self.create_input_protos(image, input_id, self.dataset_id, geo_info,
                                                   metadata)
      self._all_input_protos[input_id] = input_image_proto

      ## Iterate over each masked image and create a proto for upload to clarifai
      ## The length of masks/polygons-list and labels must be equal
      for i, _polygon in enumerate(_polygons):
        try:
          input_mask_proto = self.create_mask_proto(labels[i], _polygon, input_id, self.dataset_id)
          self._all_annotation_protos[input_id].append(input_mask_proto)
        except IndexError:
          continue



from unittest import mock
from unittest.mock import Mock

import pytest as pytest

from clarifai.auth.helper import ClarifaiAuthHelper, clear_cache


@pytest.fixture(autouse=True)
def clear_caches():
  clear_cache()


def test_ui_default_url():
  default = ClarifaiAuthHelper("clarifai", "main", "fake_pat")
  assert default.ui == "https://clarifai.com"
  assert default.pat == "fake_pat"


@pytest.mark.parametrize(("input_url", "expected_url"), (
    ("http://localhost:3002", "http://localhost:3002"),
    ("https://localhost:3002", "https://localhost:3002"),
    ("https://clarifai.com", "https://clarifai.com"),
))
def test_ui_urls(input_url, expected_url):
  helper = ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui=input_url)
  assert helper.ui == expected_url


def test_passing_no_schema_url_use_https_when_server_is_running():

  def raise_exception():
    return Mock()

  with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
    helper = ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="server")
    assert helper.ui == "https://server"


def test_passing_no_schema_url_show_error_when_not_server_running():

  def raise_exception():
    return Mock(side_effect=Exception("http_exception"))

  with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
    with pytest.raises(
        Exception,
        match="Could not get a valid response from url: localhost:3002, is the API running there?"
    ):
      ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="localhost:3002")


def test_passing_no_schema_url_detect_http_when_SSL_in_error():

  def raise_exception():
    return Mock(side_effect=Exception("Has SSL in error"))

  with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
    helper = ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="localhost:3002")
    assert helper.ui == "http://localhost:3002"


def test_passing_no_schema_url_require_port():

  def raise_exception():
    return Mock(side_effect=Exception("Has SSL in error"))

  with mock.patch('urllib.request.urlopen', new_callable=raise_exception):
    with pytest.raises(
        Exception, match="When providing an insecure url it must have both host:port format"):
      ClarifaiAuthHelper("clarifai", "main", "fake_pat", ui="localhost")


def test_exception_empty_user():
  ClarifaiAuthHelper("", "main", "fake_pat", validate=False)
  with pytest.raises(
      Exception,
      match="Need 'user_id' to not be empty in the query params or user CLARIFAI_USER_ID env var"):
    ClarifaiAuthHelper("", "main", "fake_pat")


def test_exception_empty_app():
  ClarifaiAuthHelper("clarifai", "", "fake_pat", validate=False)
  with pytest.raises(
      Exception,
      match="Need 'app_id' to not be empty in the query params or user CLARIFAI_APP_ID env var"):
    ClarifaiAuthHelper("clarifai", "", "fake_pat")


def test_exception_empty_pat():
  ClarifaiAuthHelper("clarifai", "main", "", validate=False)
  with pytest.raises(
      Exception,
      match=
      "Need 'pat' or 'token' in the query params or use one of the CLARIFAI_PAT or CLARIFAI_SESSION_TOKEN env vars"
  ):
    ClarifaiAuthHelper("clarifai", "main", "")



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface to Clarifai Models API."""

from typing import Dict, Type

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct

from clarifai.auth.helper import ClarifaiAuthHelper
from clarifai.client import create_stub


class Models:
  """
  Interface to Clarifai models api
  """

  def __init__(self, auth: Type[ClarifaiAuthHelper]) -> None:
    self.auth = auth
    self.stub = create_stub(self.auth)

  def list_model_types(self) -> Dict:
    """
    List all API Model Types that support input and output.

    Returns:
      - A dict of;
          model_type, [{expected_input_name: value}, {expected_output_name: value}, model_descripton]
        key, value pairs respectively.i.e.
          {model_type: [{inp_field_name: value}, {output_field_name: value,...}, model_description]}

      - model_type: supported input shapes and data types dict. Structure;
          {model_type: [(supported_input_N_dims, dtypes_N),...]}

      - model_type: supported output shapes and data types dict. Structure;
          {model_type: [(supported_output_N_dims, dtypes_N),...]}
    """
    ## List model types from API
    model_types_response = self.stub.ListModelTypes(
        service_pb2.ListModelTypesRequest(),  #(user_app_id=auth.user_app_id),
        metadata=self.auth.metadata)
    # model types dict structure:
    # {model_type: [{inp_field_name: value,}, {output_field_name: value,...}, desc]}
    model_types = {}
    in_dims_dtype = {}  # {model_type: [(supported_input_N_dims, dtypes_N),...]}
    out_dims_dtype = {}  # {model_type: [(supported_output_N_dims, dtypes_N),...]}
    types_dict = MessageToDict(
        model_types_response, preserving_proto_field_name=True)['model_types']
    for i in range(len(types_dict)):
      model_id = types_dict[i]['id']
      model_desc = types_dict[i]['description']
      if 'expected_output_layers' in types_dict[i].keys():
        # expected_input_layers exist for all expected_output_layers
        # hence one conditional check
        model_types[model_id] = []
        in_dims_dtype[model_id] = []
        out_dims_dtype[model_id] = []
        expected_input = types_dict[i]['expected_input_layers']
        expected_output = types_dict[i]['expected_output_layers']
        for inp in expected_input:
          if 'data_field_name' in inp.keys():
            model_types[model_id].append({inp['data_field_name']: None})
          if 'shapes' in inp.keys():
            for dim in inp['shapes']:
              if 'dims' in dim.keys():
                if 'max_dims' in dim.keys():
                  in_dims_dtype[model_id].append((dim['dims'], dim['max_dims'], dim['data_type']))
                else:
                  in_dims_dtype[model_id].append((dim['dims'], dim['data_type']))
          continue
        temp_out = {}
        for each in expected_output:
          if 'data_field_name' in each.keys():
            temp_out[each['data_field_name']] = None
          if 'shapes' in each.keys():
            for dim in each['shapes']:

              if 'dims' in dim.keys():
                out_dims_dtype[model_id].append((dim['dims'], dim['data_type']))
          else:
            continue
        model_types[model_id].append(temp_out)
        model_types[model_id].append(model_desc)

    return {
        "Model Types": model_types,
        "Input Metadata": in_dims_dtype,
        "Output Metadata": out_dims_dtype
    }

  def init_model(
      self,
      model_id: str,
      model_type: str,
      description: str = "",
  ):
    """Init a new model on Clarifai platform.

    Args:
        model_id (str): Clarifai model id
        model_type (str): Clarifai model type
        description (str, optional): a description of the model. Defaults to "".

    Returns:
        dict: Clarifai api response
    """
    user_data_object = self.auth.get_user_app_id_proto()
    post_models_response = self.stub.PostModels(
        service_pb2.PostModelsRequest(
            user_app_id=user_data_object,
            models=[resources_pb2.Model(id=model_id, notes=description,
                                        model_type_id=model_type)]),
        metadata=self.auth.metadata)

    return MessageToDict(post_models_response, preserving_proto_field_name=True)

  def post_model_version(
      self,
      model_id: str,
      model_zip_url: str,
      input: dict,
      outputs: dict,
  ):
    """Post a new version of an existing model in the Clarifai platform.

    Args:
        model_id (str): Clarifai model id
        model_zip_url (str]): url of zip of model
        model_zip_url (str): url of zip of model
        input (dict): a dict where the key is clarifai input field and the value is triton model input,
            {clarifai_input_field: triton_input_filed}.
        outputs (dict): a dict where the keys are clarifai output fields and the values are triton model outputs,
            {clarifai_output_field1: triton_output_filed1, clarifai_output_field2: triton_output_filed2,...}.

    Returns:
        dict: clarifai api response
    """
    user_data_object = self.auth.get_user_app_id_proto()

    def _parse_fields_map(x):
      """parse input, outputs to Struct"""
      _fields_map = Struct()
      _fields_map.update(x)
      return _fields_map

    input_fields_map = _parse_fields_map(input)
    output_fields_map = _parse_fields_map(outputs)
    post_model_versions = self.stub.PostModelVersions(
        service_pb2.PostModelVersionsRequest(
            user_app_id=user_data_object,
            model_id=model_id,
            model_versions=[
                resources_pb2.ModelVersion(
                    pretrained_model_config=resources_pb2.PretrainedModelConfig(
                        model_zip_url=model_zip_url,
                        input_fields_map=input_fields_map,
                        output_fields_map=output_fields_map))
            ]),
        metadata=self.auth.metadata)

    return MessageToDict(post_model_versions, preserving_proto_field_name=True)

  def upload_model(
      self,
      model_id: str,
      model_zip_url: str,
      input: dict,
      outputs: dict,
      model_type: str,
      description: str = "",
  ):
    """Doing 2 requests for initializing and creating version for a new trained model to the Clarifai platform.

    Args:
        model_id (str): Clarifai model id
        model_zip_url (str): url of zip of model
        input (dict): a dict where the key is clarifai input field and the value is triton model input,
            {clarifai_input_field: triton_input_filed}
        outputs (dict): a dict where the keys are clarifai output fields and the values are triton model outputs,
            {clarifai_output_field1: triton_output_filed1, clarifai_output_field2: triton_output_filed2,...}
        model_type (str): Clarifai model type.
        description (str, optional): a description of the model. Defaults to "".

    Returns:
        dict: Clarifai api response
    """
    init_resp = self.init_model(model_id, model_type, description)
    if init_resp["status"]["code"] != "SUCCESS":
      return init_resp
    version_resp = self.post_model_version(model_id, model_zip_url, input, outputs)

    return version_resp

  def delete_model(self, model_id: str):
    """Delete model api by model id

    Args:
        model_id (str): Clarifai model id

    Returns:
        dict: clarifai api response
    """
    user_data_object = self.auth.get_user_app_id_proto()
    delete_model_response = self.stub.DeleteModel(
        service_pb2.DeleteModelRequest(
            user_app_id=user_data_object,
            model_id=model_id,
        ),
        metadata=self.auth.metadata)

    return MessageToDict(delete_model_response, preserving_proto_field_name=True)

  def delete_model_version(self, model_id: str, version_id: str):
    """Delete specific version of model

    Args:
        model_id (str): Clarifai model id
        version_id (str): version id of model that will be removed

    Returns:
        dict: Clarifai API response
    """
    user_data_object = self.auth.get_user_app_id_proto()
    delete_model_response = self.stub.DeleteModelVersion(
        service_pb2.DeleteModelVersionRequest(
            user_app_id=user_data_object, model_id=model_id, version_id=version_id),
        metadata=self.auth.metadata)

    return MessageToDict(delete_model_response, preserving_proto_field_name=True)

  def get_model(self, model_id: str):
    """Get model by id

    Args:
        model_id (str): Clarifai model id

    Returns:
        dict: Clarifai API response
    """
    user_data_object = self.auth.get_user_app_id_proto()
    response = self.stub.GetModel(
        service_pb2.GetModelRequest(
            user_app_id=user_data_object,
            model_id=model_id,
        ),
        metadata=self.auth.metadata)

    return MessageToDict(response, preserving_proto_field_name=True)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Commandline interface for model upload utils."""
import argparse

from clarifai.auth.helper import ClarifaiAuthHelper
from clarifai.models.api import Models
from clarifai.models.model_serving.constants import MODEL_TYPES
from clarifai.models.model_serving.model_config.deploy import ClarifaiFieldsMap


def deploy(model_url,
           model_id: str = None,
           model_type: str = None,
           desc: str = "",
           update_version: bool = False):
  # init Auth from env vars
  auth = ClarifaiAuthHelper.from_env()
  # init api
  model_api = Models(auth)

  # parsing model name/type.
  # if filename having this format: <model_id>_<model-type>
  # e.i yolov5s_coco_visual-dectector
  # else user has to input model_type and model_id
  zip_filename = model_url.split('/')[-1]
  zip_filename = zip_filename.split('.')[0]

  def _parse_name(name):
    *id_, type_ = name.split('_')
    return "_".join(id_), type_

  # parse model_id
  if not model_id and "_" in zip_filename:
    model_id = _parse_name(zip_filename)[0]
  assert model_id, "Can not parse model_id from url, please input it directly"
  # parse model_type
  if not model_type and "_" in zip_filename:
    model_type = _parse_name(zip_filename)[-1]
  assert model_type, "Can not parse model_type from url, please input it directly"
  # key map
  assert model_type in MODEL_TYPES, f"model_type should be one of {MODEL_TYPES}"
  clarifai_key_map = ClarifaiFieldsMap(model_type=model_type)
  # if updating new version of existing model
  if update_version:
    resp = model_api.post_model_version(
        model_id=model_id,
        model_zip_url=model_url,
        input=clarifai_key_map.input_fields_map,
        outputs=clarifai_key_map.output_fields_map,
    )
  # creating new model
  else:
    # post model
    resp = model_api.upload_model(
        model_id=model_id,
        model_zip_url=model_url,
        model_type=model_type,
        input=clarifai_key_map.input_fields_map,
        outputs=clarifai_key_map.output_fields_map,
        description=desc,
    )
  # response
  if resp["status"]["code"] != "SUCCESS":
    raise Exception("Post models failed, details: {}, {}".format(resp["status"]["description"],
                                                                 resp["status"]["details"]))
  else:
    print("Success!")


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  # args
  parser.add_argument("--url", type=str, required=True, help="Direct download url of zip file")
  parser.add_argument("--model_id", type=str, required=False, default="", help="Custom model id.")
  parser.add_argument(
      "--model_type",
      type=str,
      required=False,
      choices=MODEL_TYPES,
      default="",
      help="Clarifai model type")
  parser.add_argument(
      "--desc", type=str, required=False, default="", help="Short desccription of model")
  parser.add_argument(
      "--update_version",
      action="store_true",
      required=False,
      help="Update exist model with new version")
  args = parser.parse_args()
  deploy(
      model_url=args.url,
      model_id=args.model_id,
      desc=args.desc,
      model_type=args.model_type,
      update_version=args.update_version)


if __name__ == "__main__":
  main()



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton Model Config classes."""

from dataclasses import dataclass, field
from typing import List


@dataclass
class InputConfig:
  """
  Triton Input definition.
  Params:
  -------
  name: input name
  data_type: input data type
  dims: Pre-defined input data shape(s).

  Returns:
  --------
  InputConfig
  """
  name: str
  data_type: str = None
  dims: List = field(default_factory=list)


@dataclass
class OutputConfig:
  """
  Triton Output definition.
  Params:
  -------
  name: output name
  data_type: output data type
  dims: Pre-defined output data shape(s).
  labels (bool): If labels file is required for inference.

  Returns:
  --------
  OutputConfig
  """
  name: str
  data_type: str = None
  dims: List = field(default_factory=list)
  labels: bool = False

  def __post_init__(self):
    if self.labels:
      self.label_filename = "labels.txt"
    else:
      del self.labels


@dataclass
class Device:
  """
  Triton instance_group.
  Define the type of inference device and number of devices to use.
  Params:
  -------
  count: number of devices
  use_gpu: whether to use cpu or gpu.

  Returns:
  --------
  Device object
  """
  count: int = 1
  use_gpu: bool = True

  def __post_init__(self):
    if self.use_gpu:
      self.kind: str = "KIND_GPU"
    else:
      self.kind: str = "KIND_CPU"


@dataclass
class DynamicBatching:
  """
  Triton dynamic_batching config.
  Params:
  -------
  preferred_batch_size: batch size
  max_queue_delay_microseconds: max queue delay for a request batch

  Returns:
  --------
  DynamicBatching object
  """
  #preferred_batch_size: List[int] = [1] # recommended not to set
  max_queue_delay_microseconds: int = 500


@dataclass
class TritonModelConfig:
  """
  Triton Model Config base.
  Params:
  -------
  name: triton inference model name
  input: a list of an InputConfig field
  output: a list of OutputConfig fields/dicts
  instance_group: Device. see Device
  dynamic_batching: Triton dynamic batching settings.
  max_batch_size: max request batch size
  backend: Triton Python Backend. Constant

  Returns:
  --------
  TritonModelConfig
  """
  model_name: str
  model_version: str
  model_type: str
  input: List[InputConfig] = field(default_factory=list)
  output: List[OutputConfig] = field(default_factory=list)
  instance_group: Device = Device()
  dynamic_batching: DynamicBatching = DynamicBatching()
  max_batch_size: int = 1
  backend: str = "python"

  def __post_init__(self):
    """
    Set supported input dims and data_types for
    a given model_type.
    """
    image_input = InputConfig(name="image", data_type="TYPE_UINT8", dims=[-1, -1, 3])
    text_input = InputConfig(name="text", data_type="TYPE_STRING", dims=[1])

    if self.model_type == "visual-detector":
      self.input.append(image_input)
      pred_bboxes = OutputConfig(name="predicted_bboxes", data_type="TYPE_FP32", dims=[-1, 4])
      pred_labels = OutputConfig(
          name="predicted_labels", data_type="TYPE_INT32", dims=[-1, 1], labels=True)
      del pred_labels.labels
      pred_scores = OutputConfig(name="predicted_scores", data_type="TYPE_FP32", dims=[-1, 1])
      self.output.extend([pred_bboxes, pred_labels, pred_scores])

    elif self.model_type == "visual-classifier":
      self.input.append(image_input)
      pred_labels = OutputConfig(
          name="softmax_predictions", data_type="TYPE_FP32", dims=[-1], labels=True)
      del pred_labels.labels
      self.output.append(pred_labels)

    elif self.model_type == "text-classifier":
      self.input.append(text_input)
      pred_labels = OutputConfig(
          name="softmax_predictions", data_type="TYPE_FP32", dims=[-1], labels=True)
      #'Len of out list expected to be the number of concepts returned by the model,
      # with each value being the confidence for the respective model output.
      del pred_labels.labels
      self.output.append(pred_labels)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface to Clarifai Models API."""

from typing import Type

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf import json_format

from clarifai.auth.helper import ClarifaiAuthHelper
from clarifai.client import create_stub


class BaseRunner:
  """
  Interface to Clarifai models api
  """

  def __init__(self, auth: Type[ClarifaiAuthHelper], runner_id: str) -> None:
    self.auth = auth
    self.stub = create_stub(self.auth)
    self.runner_id = runner_id

  def start(self):
    self._long_poll_loop()

  def run(self, item: service_pb2.PostModelOutputsRequest) -> service_pb2.MultiOutputResponse:
    """
    Run the model on the given item.
    """
    outputs = []
    for inp in item.inputs:
      outputs.append(self.run_input(inp))

    return service_pb2.MultiOutputResponse(
        status=status_pb2.Status(
            code=status_code_pb2.SUCCESS,
            description="Success",
        ),
        outputs=outputs,
    )

  def run_input(self, input: resources_pb2.Input) -> resources_pb2.Output:
    """
    Run the model on the given input.
    """
    raise NotImplementedError("run_input() not implemented")

  def _long_poll_loop(self):
    c = 0
    while True:
      # Long poll waiting for work.
      print("Loop iteration: {}".format(c))
      work_response = self.stub.ListRunnerItems(
          service_pb2.ListRunnerItemsRequest(
              user_app_id=self.auth.get_user_app_id_proto(), runner_id=self.runner_id))
      if work_response.status.code == status_code_pb2.RUNNER_NEEDS_RETRY:
        c += 1
        continue  # immediate restart the long poll
      if work_response.status.code != status_code_pb2.SUCCESS:
        raise Exception("Error getting work: {}".format(work_response.status.description))
      if len(work_response.items) == 0:
        print("No work to do. Waiting...")
        continue

      # We have work to do. Run the model on the inputs.
      for item in work_response.items:
        if not item.HasField('post_model_outputs_request'):
          raise Exception("Unexpected work item type: {}".format(item))
        print(
            f"Working on item: {item.id} with inputs {len(item.post_model_outputs_request.inputs)}"
        )
        result = self.run(item.post_model_outputs_request)

        result_response = self.stub.PostRunnerItemOutputs(
            service_pb2.PostRunnerItemOutputsRequest(
                user_app_id=self.auth.get_user_app_id_proto(),
                item_id=item.id,
                runner_id=self.runner_id,
                runner_item_outputs=[service_pb2.RunnerItemOutput(multi_output_response=result)]))
        if result_response.status.code != status_code_pb2.SUCCESS:
          raise Exception(
              json_format.MessageToJson(result_response, preserving_proto_field_name=True))
          # raise Exception("Error posting result: {}".format(result_response.status.description))



from typing import Iterator, List

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct
from tqdm import tqdm

from .base import ClarifaiDataset


class TextClassificationDataset(ClarifaiDataset):
  """
  Upload text classification datasets to clarifai datasets
  """

  def __init__(self, datagen_object: Iterator, dataset_id: str, split: str) -> None:
    super().__init__(datagen_object, dataset_id, split)
    self._extract_protos()

  def create_input_protos(self, text_input: str, labels: List[str], input_id: str, dataset_id: str,
                          metadata: Struct) -> resources_pb2.Input:
    """
    Create input protos for each text, label input pairs.
    Args:
    	`text_input`: text string.
    	`labels`: text labels
    	`input_id: unique input id
    	`dataset_id`: Clarifai dataset id
    	`metadata`:input metadata
    Returns:
    	An input proto representing a single row input
    """
    input_proto = resources_pb2.Input(
        id=input_id,
        dataset_ids=[dataset_id],
        data=resources_pb2.Data(
            text=resources_pb2.Text(raw=text_input),
            concepts=[
                resources_pb2.Concept(
                    id=f"id-{''.join(_label.split(' '))}", name=_label, value=1.)
                for _label in labels
            ],
            metadata=metadata))

    return input_proto

  def _extract_protos(self) -> None:
    """
    Creates input protos for each data generator item.
    """
    for i, item in tqdm(enumerate(self.datagen_object), desc="Loading text data"):
      metadata = Struct()
      text = item.text
      labels = item.labels if isinstance(item.labels, list) else [item.labels]  # clarifai concept
      input_id = f"{self.dataset_id}-{self.split}-{i}" if item.id is None else f"{self.split}-{str(item.id)}"
      metadata.update({"split": self.split})

      self.input_ids.append(input_id)
      input_proto = self.create_input_protos(text, labels, input_id, self.dataset_id, metadata)

      self._all_input_protos[input_id] = input_proto



# This file contains boilerplate code to allow users write their model
# inference code that will then interact with the Triton Inference Server
# Python backend to serve end user requests.
# The module name, module path, class name & get_predictions() method names MUST be maintained as is
# but other methods may be added within the class as deemed fit provided
# they are invoked within the main get_predictions() inference method
# if they play a role in any step of model inference
"""User model inference script."""

import os
from pathlib import Path


class InferenceModel:
  """User model inference class."""

  def __init__(self) -> None:
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    self.base_path: Path = os.path.dirname(__file__)
    ## sample model loading code:
    #self.checkpoint_path: Path = os.path.join(self.base_path, "your checkpoint filename/path")
    #self.model: Callable = <load_your_model_here from checkpoint or folder>

  #Add relevant model type decorator to the method below (see docs/model_types for ref.)
  def get_predictions(self, input_data):
    """
    Main model inference method.

    Args:
    -----
      input_data: A single input data item to predict on.
        Input data can be an image or text, etc depending on the model type.

    Returns:
    --------
      One of the clarifai.models.model_serving.models.output types. Refer to the README/docs
    """
    # Delete/Comment out line below and add your inference code
    raise NotImplementedError()



from collections import namedtuple

import pytest

from clarifai.urls.helper import ClarifaiUrlHelper

auth_obj = namedtuple("auth", "ui")


@pytest.fixture()
def helper():
  auth = auth_obj(ui="http://fake")
  return ClarifaiUrlHelper(auth)


def test_module_Ui_url(helper):
  url = helper.module_ui_url("clarifai", "main", "module_1", "version_1")
  assert (url == "http://fake/clarifai/main/modules/module_1/versions/version_1")


def test_module_install_url(helper):
  install_url = helper.module_install_ui_url("person", "app", "XXX")
  assert (
      install_url ==
      "http://fake/person/app/installed_module_versions/module_manager_install/install?install=XXX"
  )


def test_install_with_custom_imv_id():
  auth = auth_obj(ui="http://fake")

  custom_imv_id = "some_imv_id"
  helper = ClarifaiUrlHelper(auth, custom_imv_id)
  install_url = helper.module_install_ui_url("person", "app", "XXX")
  assert (install_url ==
          "http://fake/person/app/installed_module_versions/some_imv_id/install?install=XXX")


def test_split_of_ui_clarifai_url():
  (
      user_id,
      app_id,
      module_id,
      module_version_id,
  ) = ClarifaiUrlHelper.split_module_ui_url(
      "clarifai.com/clarifai/main/modules/module_1/versions/2")
  assert user_id == "clarifai"
  assert app_id == "main"
  assert module_id == "module_1"
  assert module_version_id == "2"


def test_split_with_https_prefix():
  (
      user_id,
      app_id,
      module_id,
      module_version_id,
  ) = ClarifaiUrlHelper.split_module_ui_url(
      "https://clarifai.com/clarifai/main/modules/module_1/versions/2")
  assert user_id == "clarifai"
  assert app_id == "main"
  assert module_id == "module_1"
  assert module_version_id == "2"


def test_too_many_things_happen():
  with pytest.raises(
      ValueError,
      match="Provided url must have 4 or 6 parts after the domain name",
  ):
    ClarifaiUrlHelper.split_module_ui_url(
        "https://clarifai.com/clarifai/main/modules/module_1/versions/2/something")


def test_not_enougth_items():
  with pytest.raises(
      ValueError, match="Provided url must have 4 or 6 parts after the domain name"):
    ClarifaiUrlHelper.split_module_ui_url("clarifai/main/modules/module_1/versions/2")


def test_no_right_modules_in_path():
  with pytest.raises(
      ValueError, match="Provided url must have 4 or 6 parts after the domain name"):
    ClarifaiUrlHelper.split_module_ui_url("clarifai/main/modules_abc/module_1/versions/2")


def test_no_right_module_version_in_path():
  with pytest.raises(
      ValueError, match="Provided url must have 4 or 6 parts after the domain name"):
    ClarifaiUrlHelper.split_module_ui_url("clarifai/main/modules/module_1/module_versions_abc/2")


def test_no_right_model_version_in_path():
  with pytest.raises(
      ValueError, match="Provided url must have 4 or 6 parts after the domain name"):
    ClarifaiUrlHelper.split_clarifai_url("clarifai/main/models/model_1/model_versions_abc/2")


def test_split_model_with_https_prefix():
  auth = auth_obj(ui="https://clarifai.com")
  url = "https://clarifai.com/clarifai/main/models/model_1/versions/2"
  (
      user_id,
      app_id,
      resource_type,
      model_id,
      model_version_id,
  ) = ClarifaiUrlHelper.split_clarifai_url(url)
  assert user_id == "clarifai"
  assert app_id == "main"
  assert resource_type == "models"
  assert model_id == "model_1"
  assert model_version_id == "2"

  new = ClarifaiUrlHelper(auth).clarifai_url(user_id, app_id, resource_type, model_id,
                                             model_version_id)
  assert new == url


def test_split_model_without_version_with_https_prefix():
  auth = auth_obj(ui="https://clarifai.com")
  url = "https://clarifai.com/clarifai/main/models/model_1"
  (
      user_id,
      app_id,
      resource_type,
      model_id,
      model_version_id,
  ) = ClarifaiUrlHelper.split_clarifai_url(url)
  assert user_id == "clarifai"
  assert app_id == "main"
  assert resource_type == "models"
  assert model_id == "model_1"
  assert model_version_id == None

  new = ClarifaiUrlHelper(auth).clarifai_url(user_id, app_id, resource_type, model_id,
                                             model_version_id)
  assert new == url


def test_bad_resource_type_in_clarifai_url():
  auth = auth_obj(ui="https://clarifai.com")
  helper = ClarifaiUrlHelper(auth)

  with pytest.raises(
      ValueError,
      match=
      "resource_type must be one of modules, models, concepts, inputs, workflows, tasks, installed_module_versions but was models_abc"
  ):
    user_id = "clarifai"
    app_id = "main"
    resource_type = "models_abc"
    model_id = "model_1"
    model_version_id = "2"
    helper.clarifai_url(user_id, app_id, resource_type, model_id, model_version_id)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton inference server Python Backend Model."""

import os
import sys

try:
  import triton_python_backend_utils as pb_utils
except ModuleNotFoundError:
  pass
from google.protobuf import text_format
from tritonclient.grpc.model_config_pb2 import ModelConfig


class TritonPythonModel:
  """
  Triton Python BE Model.
  """

  def initialize(self, args):
    """
    Triton server init.
    """
    args["model_repository"] = args["model_repository"].replace("/1/model.py", "")
    sys.path.append(os.path.dirname(__file__))
    from inference import InferenceModel

    self.inference_obj = InferenceModel()
    self.device = "cuda:0" if "GPU" in args["model_instance_kind"] else "cpu"

    # Read input_name from config file
    self.config_msg = ModelConfig()
    with open(os.path.join(args["model_repository"], "config.pbtxt"), "r") as f:
      cfg = f.read()
    text_format.Merge(cfg, self.config_msg)
    self.input_name = [inp.name for inp in self.config_msg.input][0]

  def execute(self, requests):
    """
    Serve model inference requests.
    """
    responses = []

    for request in requests:
      in_batch = pb_utils.get_input_tensor_by_name(request, self.input_name)
      in_batch = in_batch.as_numpy()
      inference_response = self.inference_obj.get_predictions(in_batch)
      responses.append(inference_response)

    return responses



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton model zip commandline interface."""

import argparse
import zipfile
from pathlib import Path
from typing import Union


def zip_dir(triton_repository_dir: Union[Path, str], zip_filename: Union[Path, str]):
  """
  Generate triton model repository zip file for upload.
  Args:
  -----
  triton_repository_dir: Directory of triton model respository to be zipped
  zip_filename: Triton model repository zip filename

  Returns:
  --------
  None
  """
  # Convert to Path object
  dir = Path(triton_repository_dir)

  with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zip_file:
    for entry in dir.rglob("*"):
      zip_file.write(entry, entry.relative_to(dir))


def main():
  """Triton model zip cli."""
  parser = argparse.ArgumentParser(__doc__)
  parser.add_argument(
      "--triton_model_repository",
      type=str,
      required=True,
      help="Path to the triton model repository to zip.")
  parser.add_argument(
      "--zipfile_name",
      type=str,
      required=True,
      help="Name of the zipfile to be created. \
      <model_name>_<model_type> is the recommended naming convention.e.g. yolov5_visual-detector.zip"
  )
  args = parser.parse_args()
  zip_dir(args.triton_model_repository, args.zipfile_name)


if __name__ == "__main__":
  main()



#! Clarifai data upload

import importlib
import inspect
import os
import sys
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import cpu_count
from typing import Iterator, List, Optional, Tuple, Union

from clarifai_grpc.grpc.api import resources_pb2, service_pb2, service_pb2_grpc
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf.json_format import MessageToDict
from tqdm import tqdm

from clarifai.client import create_stub
from clarifai.data_upload.datasets.base import Chunker
from clarifai.data_upload.datasets.image import (VisualClassificationDataset,
                                                 VisualDetectionDataset, VisualSegmentationDataset)
from clarifai.data_upload.datasets.text import TextClassificationDataset


def load_dataset(module_dir: Union[str, os.PathLike], split: str) -> Iterator:
  """
  Validate and import dataset module data generator.
  Args:
    `module_dir`: relative path to the module directory
    The directory must contain a `dataset.py` script and the data itself.
    `split`: "train" or "val"/"test" dataset split
  Module Directory Structure:
  ---------------------------
      <folder_name>/
      ├──__init__.py
      ├──<Your local dir dataset>/
      └──dataset.py
  dataset.py must implement a class named following the convention,
  <dataset_name>Dataset and this class must have a dataloader()
  generator method
  """
  sys.path.append(str(module_dir))

  if not os.path.exists(os.path.join(module_dir, "__init__.py")):
    with open(os.path.join(module_dir, "__init__.py"), "w"):
      pass

  import dataset  # dataset module

  # get main module class
  main_module_cls = None
  for name, obj in dataset.__dict__.items():
    if inspect.isclass(obj) and "Dataset" in name:
      main_module_cls = obj
    else:
      continue

  return main_module_cls(split).dataloader()


def load_zoo_dataset(name: str, split: str) -> Iterator:
  """
  Get dataset generator object from dataset zoo.
  Args:
    `name`: dataset module name in datasets/zoo/.
    `split`: "train" or "val"/"test" dataset split
  Returns:
    Data generator object
  """
  zoo_dataset = importlib.import_module(f"datasets.zoo.{name}")
  # get main module class
  main_module_cls = None
  for name, obj in zoo_dataset.__dict__.items():
    if inspect.isclass(obj) and "Dataset" in name:
      main_module_cls = obj
    else:
      continue

  return main_module_cls(split).dataloader()


class UploadConfig:

  def __init__(
      self,
      user_id: str,
      app_id: str,
      pat: str,
      dataset_id: str,
      task: str,
      from_module: Optional[Union[str, os.PathLike]] = None,
      from_zoo: Optional[str] = None,  # load dataset from zoo
      split: str = "train",  # train or test/val
      chunk_size: int = 128,
      portal: str = "clarifai"):
    """
    Initialize upload configs.
    Args:
      `user_id`: Clarifai user id.
      `app_id`: Clarifai app id.
      `pat`: Clarifai PAT(Personal Access Token).
      `dataset_id`: Clarifai dataset id (where data is to be uploaded).
      `task`: either of `visual_clf`, `visual_detection`, `visual_segmentation` or `text_clf`.
      `from_module`: Path to dataset module directory.
        Should be left as None if `from_zoo` is to be used.
      `from_zoo`: Name of dataset to upload from the zoo.
        The name must match the dataset module name excluding the file extension.
        Should be left as None if `from_module` is to be used.
      `split`: Dataset split to upload. Either of train or test/val
      `chunk_size`: size of chunks for parallel data upload.
    """
    self.USER_ID = user_id
    self.APP_ID = app_id
    self.PAT = pat
    self.dataset_id = dataset_id
    self.task = task
    self.module_dir = from_module
    self.zoo_dataset = from_zoo
    self.split = split
    self.chunk_size = min(128, chunk_size)  # limit max protos in a req
    self.num_workers: int = min(10, cpu_count())  #15 req/sec rate limit
    self.annot_num_workers = 4
    self.max_retires = 10
    self.__base: str = ""
    if portal == "dev":
      self.__base = "https://api-dev.clarifai.com"
    elif portal == "staging":
      self.__base = "https://api-staging.clarifai.com"
    else:  #prod
      self.__base = "https://api.clarifai.com"

    # Set auth vars as env variables
    os.environ["CLARIFAI_USER_ID"] = self.USER_ID
    os.environ["CLARIFAI_APP_ID"] = self.APP_ID
    os.environ["CLARIFAI_API_BASE"] = self.__base
    os.environ["CLARIFAI_PAT"] = self.PAT

    self.STUB: service_pb2_grpc.V2Stub = create_stub()
    self.metadata: Tuple = (('authorization', 'Key ' + self.PAT),)
    self.user_app_id = resources_pb2.UserAppIDSet(user_id=self.USER_ID, app_id=self.APP_ID)

  def _upload_inputs(self, batch_input: List[resources_pb2.Input]) -> str:
    """
    Upload inputs to clarifai platform dataset.
    Args:
      batch_input: input batch protos
    Returns:
      input_job_id: Upload Input Job ID
    """
    input_job_id = uuid.uuid4().hex  # generate a unique id for this job
    response = self.STUB.PostInputs(
        service_pb2.PostInputsRequest(
            user_app_id=self.user_app_id, inputs=batch_input, inputs_add_job_id=input_job_id),)
    if response.status.code != status_code_pb2.SUCCESS:
      try:
        print(f"Post inputs failed, status: {response.inputs[0].status.details}")
      except:
        print(f"Post inputs failed, status: {response.status.details}")

    return input_job_id

  def _upload_annotations(self, batch_annot: List[resources_pb2.Annotation]
                         ) -> Union[List[resources_pb2.Annotation], List[None]]:
    """
    Upload image annotations to clarifai detection dataset
    Args:
      batch_annot: annot batch protos
    Returns:
      retry_upload: failed annot upload
    """
    retry_upload = []  # those that fail to upload are stored for retries
    response = self.STUB.PostAnnotations(
        service_pb2.PostAnnotationsRequest(user_app_id=self.user_app_id, annotations=batch_annot),)

    if response.status.code != status_code_pb2.SUCCESS:
      try:
        print(f"Post annotations failed, status: {response.annotations[0].status.details}")
      except:
        print(f"Post annotations failed, status: {response.status.details}")
      finally:
        retry_upload.extend(batch_annot)

    return retry_upload

  def _concurrent_annot_upload(self, annots: List[List[resources_pb2.Annotation]]
                              ) -> Union[List[resources_pb2.Annotation], List[None]]:
    """
    Uploads annotations concurrently.
    Args:
      annots: annot protos
    Returns:
      retry_annot_upload: All failed annot protos during upload
    """
    annot_threads = []
    retry_annot_upload = []

    with ThreadPoolExecutor(max_workers=self.annot_num_workers) as executor:  # limit annot workers
      annot_threads = [
          executor.submit(self._upload_annotations, inp_batch) for inp_batch in annots
      ]

      for job in as_completed(annot_threads):
        result = job.result()
        if result:
          retry_annot_upload.extend(result)

    return retry_annot_upload

  def _backoff_iterator(self) -> None:
    """
    Return iterator for exponential backoff intervals.
    """
    yield 0.1
    for i in range(5, 11):
      yield 0.01 * (2**i)
    while True:
      yield 0.01 * (2**10)  #10 sec

  def _wait_for_inputs(self, input_job_id: str) -> bool:
    """
    Wait for inputs to be processed. Cancel Job if timeout > 30 minutes.
    Args:
      input_job_id: Upload Input Job ID
    Returns:
      True if inputs are processed, False otherwise
    """
    backoff_iterator = self._backoff_iterator()
    max_retries = self.max_retires
    start_time = time.time()
    while True:
      response = self.STUB.GetInputsAddJob(
          service_pb2.GetInputsAddJobRequest(user_app_id=self.user_app_id, id=input_job_id),)

      if time.time() - start_time > 60 * 30 or max_retries == 0:  # 30 minutes timeout
        self.STUB.CancelInputsAddJob(
            service_pb2.CancelInputsAddJobRequest(user_app_id=self.user_app_id, id=input_job_id),
        )  #Cancel Job
        return False
      if response.status.code != status_code_pb2.SUCCESS:
        max_retries -= 1
        print(f"Get input job failed, status: {response.status.details}\n")
        continue
      if response.inputs_add_job.progress.in_progress_count == 0 and response.inputs_add_job.progress.pending_count == 0:
        return True
      else:
        time.sleep(next(backoff_iterator))

  def _delete_failed_inputs(self, input_ids: List[str]) -> Tuple[List[str], List[str]]:
    """
    Delete failed input ids from clarifai platform dataset.
    Args:
      input_ids: batch input ids
    Returns:
      success_inputs: upload success input ids
      failed_inputs: upload failed input ids
    """
    success_status = status_pb2.Status(code=status_code_pb2.INPUT_DOWNLOAD_SUCCESS)
    response = self.STUB.ListInputs(
        service_pb2.ListInputsRequest(
            ids=input_ids,
            per_page=len(input_ids),
            user_app_id=self.user_app_id,
            status=success_status),)
    response_dict = MessageToDict(response)
    success_inputs = response_dict.get('inputs', [])

    success_input_ids = [input.get('id') for input in success_inputs]
    failed_input_ids = list(set(input_ids) - set(success_input_ids))
    #delete failed inputs
    self.STUB.DeleteInputs(
        service_pb2.DeleteInputsRequest(user_app_id=self.user_app_id, ids=failed_input_ids),)

    return success_input_ids, failed_input_ids

  def _upload_inputs_annotations(
      self, batch_input_ids: List[str]) -> Tuple[List[str], List[resources_pb2.Annotation]]:
    """
    Uploads batch of inputs and annotations concurrently to clarifai platform dataset.
    Args:
      batch_input_ids: batch input ids
    Returns:
      failed_input_ids: failed input ids
      retry_annot_protos: failed annot protos
    """
    input_protos, _ = self.dataset_obj.get_protos(batch_input_ids)
    input_job_id = self._upload_inputs(input_protos)
    retry_annot_protos = []

    self._wait_for_inputs(input_job_id)
    success_input_ids, failed_input_ids = self._delete_failed_inputs(batch_input_ids)

    if self.task in ["visual_detection", "visual_segmentation"]:
      _, annotation_protos = self.dataset_obj.get_protos(success_input_ids)
      chunked_annotation_protos = Chunker(annotation_protos, self.chunk_size).chunk()
      retry_annot_protos.extend(self._concurrent_annot_upload(chunked_annotation_protos))

    return failed_input_ids, retry_annot_protos

  def _retry_uploads(self, failed_input_ids: List[str],
                     retry_annot_protos: List[resources_pb2.Annotation]) -> None:
    """
    Retry failed uploads.
    Args:
      failed_input_ids: failed input ids
      retry_annot_protos: failed annot protos
    """
    if failed_input_ids:
      self._upload_inputs_annotations(failed_input_ids)
    if retry_annot_protos:
      chunked_annotation_protos = Chunker(retry_annot_protos, self.chunk_size).chunk()
      _ = self._concurrent_annot_upload(chunked_annotation_protos)

  def _data_upload(self, input_ids: List[str]) -> None:
    """
    Uploads inputs and annotations to clarifai platform dataset.
    Args:
      input_ids: input ids
    """
    chunk_input_ids = Chunker(input_ids, self.chunk_size).chunk()
    with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
      with tqdm(total=len(chunk_input_ids), desc='Uploading Dataset') as progress:
        # Submit all jobs to the executor and store the returned futures
        futures = [
            executor.submit(self._upload_inputs_annotations, batch_input_ids)
            for batch_input_ids in chunk_input_ids
        ]

        for job in as_completed(futures):
          retry_input_proto, retry_annot_protos = job.result()
          self._retry_uploads(retry_input_proto, retry_annot_protos)
          progress.update()

  def upload_to_clarifai(self):
    """
    Execute data upload.
    """
    datagen_object = None
    if self.module_dir is None and self.zoo_dataset is None:
      raise Exception("One of `from_module` and `from_zoo` must be \
      specified. Both can't be None or defined at the same time.")
    elif self.module_dir is not None and self.zoo_dataset is not None:
      raise Exception("Use either of `from_module` or `from_zoo` \
      but NOT both.")
    elif self.module_dir is not None:
      datagen_object = load_dataset(self.module_dir, self.split)
    else:
      datagen_object = load_zoo_dataset(self.zoo_dataset, self.split)

    if self.task == "text_clf":
      self.dataset_obj = TextClassificationDataset(datagen_object, self.dataset_id, self.split)
      self._data_upload(self.dataset_obj.input_ids)

    elif self.task == "visual_detection":
      self.dataset_obj = VisualDetectionDataset(datagen_object, self.dataset_id, self.split)
      self._data_upload(self.dataset_obj.input_ids)  # TODO: get_img_ids or get_input_ids

    elif self.task == "visual_segmentation":
      self.dataset_obj = VisualSegmentationDataset(datagen_object, self.dataset_id, self.split)
      self._data_upload(self.dataset_obj.input_ids)

    else:  # visual-classification & visual-captioning
      self.dataset_obj = VisualClassificationDataset(datagen_object, self.dataset_id, self.split)
      self._data_upload(self.dataset_obj.input_ids)



#! IMDB 50k Movie Reviews dataset

import csv
import os

from clarifai.data_upload.datasets.features import TextFeatures


class IMDBMovieReviewsDataset:
  """IMDB 50K Movie Reviews Dataset."""

  def __init__(self, split: str = "train"):
    """
    Initialize dataset params.
    Args:
      data_dir: the local dataset directory.
      split: "train" or "test"
    """
    self.split = split
    self.data_dirs = {
        "train": os.path.join(os.path.dirname(__file__), "train.csv"),
        "test": os.path.join(os.path.dirname(__file__), "test.csv")
    }

  def dataloader(self):
    """
    Transform text data into clarifai proto compatible
    format for upload.
    Returns:
      TextFeatures type generator.
    """
    ## Your preprocessing code here
    with open(self.data_dirs[self.split]) as _file:
      reader = csv.reader(_file)
      next(reader, None)  # skip header
      for review in reader:
        yield TextFeatures(
            text=review[0],  # text,
            labels=review[1],  # sentiment,
            id=None)



#! COCO 2017 Image Segmentation dataset

import gc
import os
from functools import reduce

import cv2
import numpy as np
from pycocotools import mask as maskUtils
from pycocotools.coco import COCO

from clarifai.data_upload.datasets.features import VisualSegmentationFeatures


class COCOSegmentationDataset:
  """COCO 2017 Image Segmentation Dataset.
  url: https://cocodataset.org/#download
  """

  def __init__(self, split: str = "train"):
    """
    Inititalize dataset params.
    Args:
      split: "train" or "test"
    """
    self.split = split
    self.image_dir = {"train": os.path.join(os.path.dirname(__file__), "images")}
    self.annotations_file = {
        "train":
            os.path.join(os.path.dirname(__file__), "annotations/instances_val2017_subset.json")
    }

  def dataloader(self):
    """
    Transform COCO 2017 segmentation dataset into clarifai proto compatible
    format to uplaod
    Returns:
      VisualSegmentationFeatures type generator.
    """
    coco = COCO(self.annotations_file[self.split])
    categories = coco.loadCats(coco.getCatIds())
    cat_id_map = {category["id"]: category["name"] for category in categories}
    cat_img_ids = {}
    for cat_id in list(cat_id_map.keys()):
      cat_img_ids[cat_id] = coco.getImgIds(catIds=[cat_id])

    img_ids = []
    for i in list(cat_img_ids.values()):
      img_ids.extend(i)

    # Get the image information for the specified image IDs
    image_info = coco.loadImgs(img_ids)
    # Extract the file names from the image information
    image_filenames = {img_id: info['file_name'] for info, img_id in zip(image_info, img_ids)}
    #get annotations for each image id
    for _id in set(img_ids):
      annots = []  # polygons
      class_names = []
      labels = [i for i in list(filter(lambda x: _id in cat_img_ids[x], cat_img_ids))]
      image_path = os.path.join(self.image_dir[self.split], image_filenames[_id])

      image_height, image_width = cv2.imread(image_path).shape[:2]
      for cat_id in labels:
        annot_ids = coco.getAnnIds(imgIds=_id, catIds=[cat_id])

        if len(annot_ids) > 0:
          img_annotations = coco.loadAnns(annot_ids)
          for ann in img_annotations:
            # get polygons
            if type(ann['segmentation']) == list:
              for seg in ann['segmentation']:
                poly = np.array(seg).reshape((int(len(seg) / 2), 2))
                poly[:, 0], poly[:, 1] = poly[:, 0] / image_width, poly[:, 1] / image_height
                annots.append(poly.tolist())  #[[x=col, y=row],...]
                class_names.append(cat_id_map[cat_id])
            else:  # seg: {"counts":[...]}
              if type(ann['segmentation']['counts']) == list:
                rle = maskUtils.frPyObjects([ann['segmentation']], image_height, image_width)
              else:
                rle = ann['segmentation']
              mask = maskUtils.decode(rle)  #binary mask
              #convert mask to polygons and add to annots
              contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
              polygons = []
              for cont in contours:
                if cont.size >= 6:
                  polygons.append(cont.astype(float).flatten().tolist())
              # store polygons in (x,y) pairs
              polygons_flattened = reduce(lambda x, y: x + y, polygons)
              del polygons
              del contours
              del mask
              gc.collect()

              polygons = np.array(polygons_flattened).reshape((int(len(polygons_flattened) / 2),
                                                               2))
              polygons[:, 0] = polygons[:, 0] / image_width
              polygons[:, 1] = polygons[:, 1] / image_height

              annots.append(polygons.tolist())  #[[x=col, y=row],...,[x=col, y=row]]
              class_names.append(cat_id_map[cat_id])
        else:  # if no annotations for given image_id-cat_id pair
          continue
      assert len(class_names) == len(annots), f"Num classes must match num annotations\
      for a single image. Found {len(class_names)} classes and {len(annots)} polygons."

      yield VisualSegmentationFeatures(image_path, class_names, annots, id=_id)



#! Food-101 image classification dataset

import os

from clarifai.data_upload.datasets.features import VisualClassificationFeatures


class Food101Dataset:
  """Food-101 Image Classification Dataset.
  url: https://data.vision.ee.ethz.ch/cvl/datasets_extra/food-101/
  """

  def __init__(self, split: str = "train"):
    """
    Initialize dataset params.
    Args:
      data_dir: the local dataset directory.
      split: "train" or "test"
    """
    self.split = split
    self.image_dir = {"train": os.path.join(os.path.dirname(__file__), "images")}

  def dataloader(self):
    """
    Transform food-101 dataset into clarifai proto compatible
    format for upload.
    Returns:
      VisualClassificationFeatures type generator.
    """
    ## Your preprocessing code here
    class_names = os.listdir(self.image_dir[self.split])
    for class_name in class_names:
      for image in os.listdir(os.path.join(self.image_dir[self.split], class_name)):
        image_path = os.path.join(self.image_dir[self.split], class_name, image)
        yield VisualClassificationFeatures(
            image_path=image_path,
            label=class_name,
            id=None  # or image_id
        )



# This file contains boilerplate code to allow users write their model
# inference code that will then interact with the Triton Inference Server
# Python backend to serve end user requests.
# The module name, module path, class name & get_predictions() method names MUST be maintained as is
# but other methods may be added within the class as deemed fit provided
# they are invoked within the main get_predictions() inference method
# if they play a role in any step of model inference
"""User model inference script."""

import os
from pathlib import Path
from typing import Callable

import numpy as np
import torch

from clarifai.models.model_serving.models.model_types import visual_detector
from clarifai.models.model_serving.models.output import VisualDetectorOutput


class InferenceModel:
  """User model inference class."""

  def __init__(self) -> None:
    """
    Load inference time artifacts that are called frequently .e.g. models, tokenizers, etc.
    in this method so they are loaded only once for faster inference.
    """
    self.base_path: Path = os.path.dirname(__file__)
    self.checkpoint_path = os.path.join(self.base_path, "model.pt")  #yolov5x
    self.model: Callable = torch.hub.load(
        os.path.join(self.base_path, 'yolov5'),
        'custom',
        autoshape=True,
        path=self.checkpoint_path,
        source='local')
    self.device = "cuda:0" if torch.cuda.is_available() else "cpu"

  @visual_detector
  def get_predictions(self, input_data) -> VisualDetectorOutput:
    """
    Main model inference method.

    Args:
    -----
      input_data: A single input data item to predict on.
        Input data can be an image or text, etc depending on the model type.

    Returns:
    --------
      One of the clarifai.models.model_serving.models.output types. Refer to the README/docs
    """
    preds = self.model(input_data)
    max_bbox_count = 300  # max allowed detected bounding boxes per image
    preds = preds.xyxy[0].cpu().numpy()
    labels = [[pred[5]] for pred in preds]
    scores = [[pred[4]] for pred in preds]
    h, w, _ = input_data.shape  # input image shape
    bboxes = [[x[1] / h, x[0] / w, x[3] / h, x[2] / w]
              for x in preds]  # normalize the bboxes to [0,1]
    if len(bboxes) != 0:
      bboxes = np.concatenate((bboxes, np.zeros((max_bbox_count - len(bboxes), 4))))
      scores = np.concatenate((scores, np.zeros((max_bbox_count - len(scores), 1))))
      labels = np.concatenate((labels, np.zeros((max_bbox_count - len(labels), 1),
                                                dtype=np.int32)))
    else:
      bboxes = np.zeros((max_bbox_count, 4), dtype=np.float32)
      scores = np.zeros((max_bbox_count, 1), dtype=np.float32)
      labels = np.zeros((max_bbox_count, 1), dtype=np.int32)

    return VisualDetectorOutput(
        predicted_bboxes=bboxes, predicted_labels=labels, predicted_scores=scores)



#! dataset output features (output from preprocessing & input to clarifai data proto builders)
from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass
class TextFeatures:
  """
  Text classification datasets preprocessing output features.
  """
  text: str
  labels: List[Union[str, int]]  # List[str or int] to cater for multi-class tasks
  id: Optional[int] = None  # text_id


@dataclass
class VisualClassificationFeatures:
  """
  Image classification datasets preprocessing output features.
  """
  image_path: str
  label: Union[str, int]
  geo_info: Optional[List[float]] = None  #[Longitude, Latitude]
  id: Optional[int] = None  # image_id


@dataclass
class VisualDetectionFeatures:
  """
  Image Detection datasets preprocessing output features.
  """
  image_path: str
  classes: List[Union[str, int]]
  bboxes: List[List[float]]
  geo_info: Optional[List[float]] = None  #[Longitude, Latitude]
  id: Optional[int] = None  # image_id


@dataclass
class VisualSegmentationFeatures:
  """
  Image Segmentation datasets preprocessing output features.
  """
  image_path: str
  classes: List[Union[str, int]]
  polygons: List[List[List[float]]]
  geo_info: Optional[List[float]] = None  #[Longitude, Latitude]
  id: Optional[int] = None  # image_id
